#include "RxslFunction.hpp"
#include "SXSLTUtils.hpp"

#include "RSCommon.h"

#undef error

#include <stdlib.h>

USER_OBJECT_ asSSymbol(const char *tmp);
USER_OBJECT_ createNodeReference(const XObjectPtr &node);

XObjectPtr
FunctionRCall::execute(
      XPathExecutionContext& executionContext,
      XalanNode* context,
      int  /* opPos */,
      const XObjectArgVectorType&    args)
{
    // Use the XObjectFactory createNumber() method to create an XObject 
    // corresponding to the XSLT number data type.
  
    init_R();

#ifdef RXSL_DEBUG
cerr << "Function name: '" << functionName() << "'" << endl;
for(int j = 0; j < args.size(); j++)
  cerr << "\t" << j << ") " << args[j]->getType() << endl;
#endif

    USER_OBJECT_ fun = Rf_findFun(Rf_install((char *)functionName()),  R_GlobalEnv);
    PROTECT(fun);

    int numArgs = args.size();
    USER_OBJECT_ e = allocVector(LANGSXP, numArgs+1);
    PROTECT(e);
    SETCAR(e, fun);

    USER_OBJECT_ tmp = CDR(e);
    for(int i = 0; i < numArgs;i++) {
      USER_OBJECT_ tmpArg = convert(args[i]);
      if(args[i]->getType() == XObject::eTypeNodeSet) {
        SETCAR(tmp, VECTOR_ELT(tmpArg, 0));
      } else
        SETCAR(tmp, tmpArg);
      tmp = CDR(tmp);
    }
      
    USER_OBJECT_ sval = eval(e, R_GlobalEnv);

    XObjectPtr ans = convert(sval, executionContext);
    UNPROTECT(2);
    return ans;
}


/*
 This is the version that gets called with no arguments from the XSL.
 Therefore, we just lookup the function and invoke it with no arguments.
*/
XObjectPtr
FunctionRCall::execute(XPathExecutionContext&			executionContext,
    	               XalanNode*						context)
{
    USER_OBJECT_ e = allocVector(LANGSXP, 1);
    PROTECT(e);
    USER_OBJECT_ fun = Rf_findFun(Rf_install((char *)functionName()),  R_GlobalEnv);
    SETCAR(e, fun);
      
    USER_OBJECT_ sval = eval(e, R_GlobalEnv);

    XObjectPtr ans = convert(sval, executionContext);
    UNPROTECT(1);
    return ans;

}

USER_OBJECT_
FunctionRCall::convert(const XObjectPtr &arg)
{
    USER_OBJECT_ val;
    switch(arg->getType()) {
      case XObject::eTypeNumber:
         PROTECT(val = NEW_NUMERIC(1));
         NUMERIC_DATA(val)[0] = arg->num(); 
         UNPROTECT(1);
       break;
      case XObject::eTypeBoolean:
         PROTECT(val = NEW_LOGICAL(1));
         LOGICAL_DATA(val)[0] = arg->boolean(); 
         UNPROTECT(1);
       break;
      case XObject::eTypeString:
         val = fromXSLString(arg);
       break;
      case XObject::eTypeNodeSet:
	{
#if 0
          /* This is a deep(ish) copy of the node. */ 
         const NodeRefListBase& nodes = arg->nodeset();
         int n = nodes.getLength();
         PROTECT(val = NEW_LIST(n));
         for(int i = 0; i < n ; i++) {
          XalanNode *tmpNode =  nodes.item(i);
           SET_VECTOR_ELT(val, i, convert(tmpNode));
	 }
#else
           /* This passes the node across as a reference. */
           val = createNodeReference(arg);
#endif
        }
       break;
      default:
        fprintf(stderr, "Unhandled case (%d) in converting to S\n", arg->getType());
        fflush(stderr);
       break;
    }

  return(val);
}


USER_OBJECT_
createNodeReference(const XObjectPtr &node)
{
  USER_OBJECT_ ans, tmp;
  PROTECT(ans = NEW_LIST(2));

   SET_VECTOR_ELT(ans, 0, tmp = NEW_NUMERIC(1));
   NUMERIC_DATA(tmp)[0] = (double) ((long int) &node);

   PROTECT(tmp = NEW_CHARACTER(1));
   SET_STRING_ELT(tmp, 0, COPY_TO_USER_STRING("XSLNodeReference"));
   SET_CLASS(ans, tmp);

   UNPROTECT(2);
   return(ans);
}

USER_OBJECT_
fromXSLString(XObjectPtr arg)
{
  USER_OBJECT_ val;
   int len = arg->str().length();
   char *tmp = getString(arg->str());
   
   double dval;
   long lval;
  
   if( ( val = asSSymbol(tmp)) != NULL) {
      val = Rf_install(tmp);      
   } else if(isXSLInteger(tmp, &lval)) {
     val = NEW_INTEGER(1);           
     INTEGER_DATA(val)[0] = lval; 
   } else if(isXSLReal(tmp, &dval)) {
       val = NEW_NUMERIC(1);           
       NUMERIC_DATA(val)[0] = dval;
   } else {
       PROTECT(val = NEW_CHARACTER(1));
       SET_STRING_ELT(val, 0, COPY_TO_USER_STRING(tmp));      
       UNPROTECT(1);
   }

 return(val);
}

USER_OBJECT_
FunctionRCall::convert(const XalanNode *node)
{
 USER_OBJECT_ sval = NULL_USER_OBJECT;
 int type = node->getNodeType();
 switch(type) {
   case XalanNode::TEXT_NODE:
   case XalanNode::CDATA_SECTION_NODE:
   case XalanNode::ATTRIBUTE_NODE:
   case XalanNode::PROCESSING_INSTRUCTION_NODE:
   {
     PROTECT(sval = NEW_CHARACTER(1));
     char *tmp = getString(node->getNodeValue());
     SET_STRING_ELT(sval, 0, COPY_TO_USER_STRING(tmp));
     UNPROTECT(1);
   }
     break;
#if 1
   case XalanNode::ELEMENT_NODE:
    {
      if(node->hasChildNodes()) {
        const XalanNode *el = node->getFirstChild();
        int n = 0;
        while(el) {
          n++;
          el = el->getNextSibling();
	}
        el = node->getFirstChild();
        PROTECT(sval = NEW_LIST(n));
        for(int i = 0; i < n;i++) {
          SET_VECTOR_ELT(sval, i,  convert(el));
          el = el->getNextSibling();
	}
        UNPROTECT(1);
#if 0
         /* This crashes for some reason. */
        const XalanNodeList *els = node->getChildNodes();
        PROTECT(sval = NEW_LIST(n));
        for(int i = 0; i < n ; i++) {
          SET_VECTOR_ELT(sval, i,  convert(els->item(i)));
        }
        UNPROTECT(1);
#endif
      } else {
 fprintf(stderr, "Have to think\n");
      }
    }
    break;
#endif
    default: 
       fprintf(stderr, "Unhandled case in convert(XalanNode*) %d\n", type); 
        fflush(stderr);
     break;
 }

 return(sval);
}

XObjectPtr
FunctionRCall::convert(USER_OBJECT_ sval, XPathExecutionContext& executionContext) 
{
    XObjectPtr ans;
//Rf_PrintValue(sval);
    switch(TYPEOF(sval)) {
        case STRSXP:
	  {
           XalanDOMString str(CHAR_DEREF(STRING_ELT(sval,0)));
           ans = executionContext.getXObjectFactory().createString(str);
          }
         break;
        case INTSXP:
          ans = executionContext.getXObjectFactory().createNumber(INTEGER_DATA(sval)[0]);
         break;
        case REALSXP:
          ans = executionContext.getXObjectFactory().createNumber(NUMERIC_DATA(sval)[0]);
         break;
        case LGLSXP:
          ans = executionContext.getXObjectFactory().createBoolean(LOGICAL_DATA(sval)[0]);
         break;
        default:
            ans = convertXMLNode(sval, executionContext);
          break;
    }

    return(ans);
}

/*
 Create a result fragment of XalanNode objects from the S representation
 of XML nodes (i.e. a top-level XMLNode object).
*/
XObjectPtr
FunctionRCall::convertXMLNode(USER_OBJECT_ sval, XPathExecutionContext &executionContext)
{
 XObjectPtr ans;

   if(!R_IS(sval, "XMLNode")) {
     // ans = createXMLNode(sval);
ans = XObjectPtr(NULL);
   }

   if(ans == (XObjectPtr) NULL || GET_LENGTH(sval) > 0)
     return( (XObjectPtr) NULL );


 XalanDocument *doc = executionContext.getDOMFactory();

#if 0
 XResultTreeFrag *top;
    
 top = new ResultTreeFrag(executionContext.getDOMFactory());
 XalanNode *node;

  top->appendChild(node);
  return(top);
#endif

#if 0
 char *foo = "Duncan";
  XObject *tmp = executionContext.startElement(foo);
  return(tmp);
#endif
}

bool
isXSLInteger(char *tmp, long *lval)
{
 bool ok = false;

 ok = tmp[0] >= '0' &&   tmp[0] <= ('0'+9);

 if(ok == false)
   return(false);

 *lval = strtol(tmp, NULL, 0); 
 if(errno)
   return(false);

 return(true);
}


bool
isXSLReal(char *tmp, double *dval)
{
 bool ok = false;

 ok = tmp[0] >= '0' &&   tmp[0] <= ('0'+9);

 if(ok == false)
   return(false);

 *dval = strtod(tmp, NULL); 
 if(errno)
   return(false);

 return(true);
}


USER_OBJECT_
asSSymbol(const char *tmp)
{
 USER_OBJECT_ val = NULL;
 USER_OBJECT_ fun;
 USER_OBJECT_ name, e, ok;
 PROTECT(fun = Rf_findFun(Rf_install("exists"), R_GlobalEnv));
 PROTECT(name = NEW_CHARACTER(1));
  SET_STRING_ELT(name, 0, COPY_TO_USER_STRING(tmp));
 PROTECT(e = allocVector(LANGSXP,2));
 SETCAR(e, fun);
 SETCAR(CDR(e), name);
 ok = eval(e, R_GlobalEnv);
 if(LOGICAL_DATA(ok)[0] == TRUE) {
  val = Rf_install((char *)tmp);
 }
 UNPROTECT(3);

 return(val);
}
